/*------------------------------------------------------------------------------------------
 																						
  Project:			Sprite Engine

  Module:			ddtools.cpp - Tools for DirectDraw.

  Author:			Matthias Hofmann ( unless otherwise stated in the function description )										
																						
--------------------------------------------------------------------------------------------*/


#include "master.h"


//
// Globals
//
LPDIRECTDRAW7					lpDD7;
LPDIRECTDRAWSURFACE7			Primary;
LPDIRECTDRAWSURFACE7			Back;


/*
 * Description:		Initializes DirectDraw and creates a flipper in videomemory. In case the 
 *					chosen color depth doesnt fit into video memory, the function decreases
 *					the color depth until it fits. The final color depth is returned. In the 
 *					unlikely event that none was found, this can be 0.
 *
 * Parameters:		hWnd - Handle to the application window.
 *					lplpDD7 - Address of a pointer to an IDirectDraw7 interface.
 *					lpsi - Pointer to a SSettingsInfo structure.
 *
 * Return value:	The color depth that was used to create the flipper, or 0 if none was found.
 *					In case of an unrecoverable error, the corresponding error code is returned.
 *
 */
int CreateFlipper( HWND hWnd, LPDIRECTDRAW7* lplpDD7, const SSettingsInfo* lpsi )
{
	DDSURFACEDESC2	ddsd;
	DDSCAPS2		ddscaps;
	int				iBitsPixel;
	int				iBkBufferCount = BACKBUFFERCOUNT;
	HRESULT			hr;

	
	// The original values in lpsi shouldnt be changed,
	// so copy them into a variable to work with.
	iBitsPixel = lpsi->bDepth;

	// Create the IDirectDraw7 interface.
	DirectDrawCreateEx( lpsi->lpGuid, ( void** ) lplpDD7, IID_IDirectDraw7, NULL );

	// Set the cooperative level.
	( *lplpDD7 )->SetCooperativeLevel( hWnd, DDSCL_EXCLUSIVE | 
											 DDSCL_FULLSCREEN | 
											 DDSCL_ALLOWREBOOT );


	//
	// Try to create a flipper until the color depth fits.
	//
	while ( iBitsPixel )
	{
		// Set the display mode to the current try.
		hr = ( *lplpDD7 )->SetDisplayMode( 800, 600, iBitsPixel, 0, 0 );


		// If this color depth is not supported,
		// try a lower one.
		if ( hr == DDERR_INVALIDMODE )
		{
			iBitsPixel -= 8;
			continue;
		}

		// Setup the surface description for the flipper.
		ZeroMemory( &ddsd, sizeof ddsd );
		ddsd.dwSize = sizeof ddsd;
		ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | 
							  DDSCAPS_FLIP | 
							  DDSCAPS_COMPLEX;
		ddsd.dwBackBufferCount = iBkBufferCount;

		// See if the current color depth works ...
		hr = ( *lplpDD7 )->CreateSurface( &ddsd, &Primary, NULL );
			
		// ... if yes, thats it ...
		if ( hr == DD_OK )
			break;

		// ... but if it doesnt fit into video memory, try lower requirements ...
		if ( hr == DDERR_OUTOFVIDEOMEMORY )
		{
			// First, try fewer back buffers ...
			if ( iBkBufferCount > 1 )
			{	  
				iBkBufferCount--;	
			}
			// ... then, reduce the color depth.
			else
			{
				iBitsPixel -= 8;
			}
		}
		// ... or report an error if something else fails.
		else
		{
			// Return the error.
			return hr;
		}
	}

	//
	// In case we found a suitable mode, get the backbuffer.
	//
	if ( iBitsPixel )
	{
		ZeroMemory( &ddscaps, sizeof ddscaps );
		ddscaps.dwCaps = DDSCAPS_BACKBUFFER;

		Primary->GetAttachedSurface( &ddscaps, &Back );
	}

	// Return the mode that was used, or 0 if none was found.
	return iBitsPixel;
}

/*
 * Description:		Creates a DirectDraw surface or restores an existing one.
 *
 * Parameters:		lplpDDS   - Address of a pointer to a DirectDraw surface. If the
 *							    pointer contains NULL, the surface will be newly created,
 *						        otherwise the surface it points at will be restored.
 *
 *					lpSize    - Pointer to a size structure that contains the desired size
 *						        of the surface. If *lplpDDS is non-zero, this parameter 
 *							    will be ignored.
 *
 *					gSysMem   - If true, the surface will be created in system memory.
 *
 * Return value:	At the moment the function always returns 0.
 *
 */
int CreateDDSurface( LPDIRECTDRAWSURFACE7* lplpDDS, LPSIZE lpSize, bool gSysMem )
{
	DDSURFACEDESC2	ddsd;


	// If the pointer referenced by lplpDDS is not pointing at NULL,
	// the surface it points at shall be restored.
	if ( *lplpDDS != NULL )
	{
		//
		// Restore the surface in case it is lost.
		//
		if ( ( *lplpDDS )->IsLost() == DDERR_SURFACELOST )
			( *lplpDDS )->Restore();
	}
	// If the pointer referenced by lplpDDS is pointing at NULL,
	// the surface shall be newly created.
	else
	{
		//
		// Initialize the surface description. A color key is always used.
		//
		ZeroMemory( &ddsd, sizeof ddsd );
		ddsd.dwSize = sizeof ddsd;	
		ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_CKSRCBLT; 
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;    
		ddsd.dwWidth  = lpSize->cx;
		ddsd.dwHeight = lpSize->cy;
		ddsd.ddckCKSrcBlt.dwColorSpaceHighValue = 0;
		ddsd.ddckCKSrcBlt.dwColorSpaceLowValue = 0;

		//
		// If desired, arrange for the surface to get into system memory
		//
		if ( gSysMem )
			ddsd.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;

		// Create the surface.
		lpDD7->CreateSurface( &ddsd, lplpDDS, NULL );
	}

	return 0;
}

/*
 * Description:		Retrieves the width and height of a bitmap file.
 *
 * Parameters:		szName - The name of the bitmap without the file extension.
 *					lpSize - Pointer to a SIZE structure to be filled with the information.
 *
 * Return value:	At the moment the function always returns 0.
 *
 */
int GetPictureSize( char* szName, LPSIZE lpSize )
{
	HBITMAP		hbm;
	BITMAP		bm;
	char		szFileName[MAX_PATH];
	
	//
	// Create the filename of the bitmap.
	//
	strcpy( szFileName, szName );
	strcat( szFileName, ".bmp" );

	// Load the image.
	hbm = ( HBITMAP ) LoadImage( NULL, szFileName, IMAGE_BITMAP, 
		0,	0, LR_LOADFROMFILE | LR_CREATEDIBSECTION );

	// Get information about the bitmap.
	GetObject( hbm, sizeof bm, &bm );

	// Return the required information.
	lpSize->cx = bm.bmWidth;
	lpSize->cy = bm.bmHeight;

	// Free the memory associated with the bitmap.
	DeleteObject( hbm );

	return 0;
}

/*
 * Description:		Loads a BMP into a surface.
 *
 * Parameters:		lpDDS - Pointer to the surface to be loaded.
 *					szName - The name of the bitmap without the file extension.
 *
 * Return value:	At the moment the function always returns 0.
 *
 */
int LoadPicture( LPDIRECTDRAWSURFACE7 lpDDS, char* szName )
{
	HBITMAP         hbm;
	HDC             hdcImage = NULL;
	HDC             hdcSurf  = NULL;
	DDSURFACEDESC2	ddsd;
	char			szFileName[MAX_PATH];
	HRESULT			hr;


	//
	// Create the filename of the bitmap.
	//
	strcpy( szFileName, szName );
	strcat( szFileName, ".bmp" );

	// Initialize surface description structure.
	ZeroMemory( &ddsd, sizeof ddsd );
	ddsd.dwSize = sizeof ddsd;

	// Get the surface description.
	lpDDS->GetSurfaceDesc( &ddsd );

	// Load the image.
	hbm = ( HBITMAP ) LoadImage( NULL, szFileName, IMAGE_BITMAP, 
		0, 0, LR_LOADFROMFILE | LR_CREATEDIBSECTION );
 
	// Create a DC thats compatible with the applications current screen ... 
	hdcImage = CreateCompatibleDC( NULL );
	
	// ... and select the image into it.
	SelectObject( hdcImage, hbm );

	// Get a DC for the surface.
	lpDDS->GetDC( &hdcSurf );

	// The BitBlt will perform format conversion as necessary.
	hr = BitBlt( hdcSurf, 0, 0, ddsd.dwWidth, ddsd.dwHeight, hdcImage, 0, 0, SRCCOPY );

	// Release the DC of the surface.
	lpDDS->ReleaseDC( hdcSurf );
		
	// Delete the compatible DC.
	DeleteDC( hdcImage );

	// Free the memory allocated for the bitmap.
	DeleteObject( hbm );

	return 0;
}

/*
 * Description:		Initializes a DirectDraw surface. If the surface doesnt exist yet,
 *					it will be created and filled from a picture file. If the surface 
 *					already exits, the surface will be restored and reloaded from a 
 *					picture file.
 *
 * Parameters:		lplpDDS   - Address of a pointer to a DirectDraw surface. If the pointer
 *							    points at NULL, the surface will be newly created, otherwise
 *							    it will be restored if necessary.
 *
 *					szName    - The name of the picture file, without the file extension.
 *			
 *					gSysMem   - If true, the surface will be created in system memory.
 *								This parameter is ignored if the surface is just being
 *								recreated.
 *
 * Return value:	If the surface was newly created, the function returns IDDS_CREATED, if
 *					the surface was merely restored it returns IDDS_RESTORED. In case the
 *					surface existed already before the call and did not need to be restored,
 *					the function returns IDDS_NOTLOST.
 *
 */
int InitDDSurface( LPDIRECTDRAWSURFACE7* lplpDDS, char* szName, bool gSysMem )
{
	SIZE	size;


	// If the pointer referenced by lplpDDS is not pointing at NULL,
	// the surface it points at shall be restored and reloaded.
	if ( *lplpDDS != NULL )
	{
		// If the surface is lost ...
		if ( ( *lplpDDS )->IsLost() == DDERR_SURFACELOST )
		{
			// ... restore it ...
			( *lplpDDS )->Restore();

			// ... and reload it.
			LoadPicture( *lplpDDS, szName );

			return IDDS_RESTORED;
		}	

		return IDDS_NOTLOST;
	}
	// If the pointer referenced by lplpDDS is pointing at NULL,
	// the surface shall be newly created and loaded.
	else
	{
		// Retrieve the size of the picture file.
		GetPictureSize( szName, &size );

		// Create the surface according to the requirements.
		CreateDDSurface( lplpDDS, &size, gSysMem );

		// Load the surface with the picture file.
		LoadPicture( *lplpDDS, szName );

		return IDDS_CREATED;
	}
}

/*
 * Description:		Displays the current frame rate. For the function to
 *					work properly, it must be called in every frame.
 *
 * Parameters:		None.
 *
 * Return value:	At the moment the function always returns 0.
 *
 */
int ShowFrameRate( void )
{	 
	HDC				hdc;
	static char		szBuffer[100];
	static DWORD	dwTime;
	static int		iFramecount = 0;

	// Count every frame ( this function must 
	// be called in every frame ).
	iFramecount++;

	// If another second has passed ...
	if ( timeGetTime() - dwTime > 999 )
	{
		// ... write the frame rate to the string buffer, ...
		wsprintf( szBuffer, "FPS: %3i", iFramecount );

		// ... reset the frame counter ...
		iFramecount = 0;

		// ... and the timer.
		dwTime = timeGetTime();
	}

	//
	// Display the frame rate.
	//
	osWork->GetDC( &hdc );
	SetTextAlign( hdc, TA_TOP | TA_RIGHT );
	SetBkMode( hdc, TRANSPARENT );
	SelectObject( hdc, g_hFont );
	SetTextColor( hdc, RGB( 0, 0, 128 ) );
	TextOut( hdc, 790, 5, szBuffer, strlen( szBuffer ) );
	osWork->ReleaseDC( hdc );

	return 0;
}
